import logging
import time
from typing import Text, Dict, Any, Optional, List, Tuple

from sqlalchemy.orm import Session

from rasax.community import config
from rasax.community.database import (
    NluEvaluation,
    NluEvaluationPrediction, Model)

from rasax.community.services.model_service import StackModelService
from rasax.community.services.stack_service import StackService

logger = logging.getLogger(__name__)

MODEL_EVALUATION_TAG = "evaluation"


class EvaluationService:

    def __init__(self,
                 stack_service: 'StackService',
                 session: Session):
        self.stack_service = stack_service
        self.session = session

    def persist_evaluation(self, project_id: Text, model: Text,
                           result: Dict[Text, Any]) -> NluEvaluation:
        intent_eval = result.get("intent_evaluation", {})

        evaluation = NluEvaluation(model_id=model,
                                   report=intent_eval.get("report"),
                                   accuracy=intent_eval.get("accuracy", 0),
                                   f1=intent_eval.get("f1_score", 0),
                                   precision=intent_eval.get("precision", 0),
                                   timestamp=time.time())

        # delete possible old evaluations
        self.delete_evaluation(project_id, model)

        self.session.add(evaluation)
        self.session.flush()  # flush to get the insertion id

        predictions = [NluEvaluationPrediction(text=p["text"],
                                               intent=p["intent"],
                                               predicted=p["predicted"],
                                               confidence=p.get("confidence",
                                                                0),
                                               evaluation_id=evaluation.id)
                       for p in intent_eval.get("predictions", [])]

        self.session.bulk_save_objects(predictions)
        self.session.commit()

        return evaluation

    def evaluation_for_model(self, project_id: Text, model: Text
                             ) -> Optional[Dict[Text, Any]]:
        evaluation = self.session.query(NluEvaluation).filter(
            NluEvaluation.model_id == model).join(Model).filter(
            Model.project_id == project_id
        ).join(NluEvaluationPrediction).first()

        if evaluation:
            return evaluation.as_dict()
        else:
            return None

    def delete_evaluation(self, project_id: Text, model: Text) -> bool:
        to_delete = self.session.query(NluEvaluation).filter(
            NluEvaluation.model_id == model).join(
            Model).filter(Model.project_id == project_id).all()

        for e in to_delete:
            self.session.delete(e)

        deleted = self.session.deleted

        self.session.commit()

        return deleted

    def formatted_evaluations(self, project_id: Text) -> List[Dict[Text, Any]]:
        results = self.evaluations(project_id)
        return [
            {
                "intent_evaluation": r["intent_evaluation"],
                "project_id": project_id,
                "model": r["model"]
            } for r in results
        ]

    def evaluations(self, project_id: Text) -> List[Dict[Text, Any]]:
        evaluations = self.session.query(NluEvaluation).join(Model).filter(
            Model.project_id == project_id).all()

        return [e.as_dict() for e in evaluations]

    async def evaluate(self, training_data: List[Dict[Text, Any]], model_name: Text
                       ) -> Tuple[Optional[Dict], int]:
        model_service = StackModelService(config.rasa_model_dir, self.session,
                                          "worker")

        # Tag the model for the evaluation so that the stack worker can pull it
        _ = model_service.tag_model(config.project_name, model_name,
                                    MODEL_EVALUATION_TAG)

        model_pull_path = StackModelService.get_model_server_url(MODEL_EVALUATION_TAG)
        results = await self.stack_service.evaluate_intents(training_data,
                                                            model_pull_path)

        model_service.delete_tag(config.project_name, model_name, MODEL_EVALUATION_TAG)

        return results
